/*
  getpat.c by Chris Spear

  v1.2  05/18/01  Strip underscores and truncate input numbers
  v1.1  03/07/01  Support negative numbers, too bad tf_strdelputp doesn't
  v1.0  02/10/01  Support Verilog-XL
  v0.9  02/05/01  Cleanup before code freeze, MTI updates
  v0.8  01/18/01  Detected too much or too little un-packed data
  v0.7  01/17/01  Detected too much or too little packed data
  v0.6  01/09/01  Keep track of the line number, return it from _next
  v0.5	01/05/01  Check for null file names in $getpat_open
  v0.4	01/03/01  Allow signal list in $getpat_next
  v0.3		  Remove limit on number of signals
  v0.2		  Add bin/hex/dec, space/packed	 
  v0.1		  Just handle values <= 32 bits, constant filename,
		  no bit subscripts or memory locations

  This has three Verilog system tasks:
     filein = $getpat_open("filename", r1, r2, r3,...);
     r = $getpat_next(filein);
     $getpat_close(filein);

  See README.txt for more information

  Use plusarg +getpat_verbose for debugging.
  If $getpat_open can't open a file, it returns -1
  If $getpat_next hits EOF or any error, it returns -1,
  Otherwise it return the current line number.
  $getpat_next skips // comments in column 1

*/

#ifdef MTI
#include "d:\progra~1\modeltech_5.4e\include\veriuser.h" /* Ugly for my PC */
#else
#include "veriuser.h"         /* TF definitions, works in VCS 5.2 & later  */
#endif
#include <stdlib.h>           /* ANSI C standard library                   */
#include <stdio.h>            /* ANSI C standard input/output library      */
#include <string.h>	      /* Added to reduce error messages		   */

#define RETURN_VAL  0
#define FILE_NAME   1         /* First arg of $getpat_open is file name    */
#define FIRST_REG   2         /* First reg in $getpat_open is file name    */
#define FILE_ARG    1         /* File pointer is first arg of _next, close */
#define MAX_LINE    2000      /* Maximum width of a line                   */
#define MAX_ARGS    2000      /* Maximum number of arguments               */

/* Linked list of signals from $getpat_open */
typedef struct t_sig
{
  int width;                  /* Width of this signal in bits              */
  int digits;                 /* Number of digits in signal: bin & hex only*/
  char fmt[30];		      /* Holds format string                       */
  struct t_sig *next_sig;     /* Next signal                               */
} s_sig, *p_sig;

typedef struct t_patfile
{
  FILE *in_file;              /* File stream for this pattern file         */
  int nump;                   /* Number of arguments in $getpat_open       */
  int line_num;               /* Line number of where we are in the file   */
  char *instance;             /* Instance pointer to $getpat_open          */
  int first_reg;              /* 2 or 3 - index of first reg in _open      */
  bool first_read_done;       /* True until first read has been done       */
  char radix;                 /* b,d,h for radix                           */
  char separator;             /* Sep. between data values, space or null   */
  struct t_sig *head_sig, *tail_sig; /* List of signals from $getpat_open  */
} s_patfile, *p_patfile;

bool gp_verbose = FALSE;         /* Flag for debug output                     */
bool ignore_truncate = FALSE; /* Ignore data truncation - NOT!             */



/**********************************************************************
 * $getpat_open check routine
 *********************************************************************/
int getpat_open_check(){
  int i, first_reg,
    nump = tf_nump();

  if (mc_scan_plusargs("getpat_verbose")) gp_verbose = TRUE; /* set verbose flag */

  /* Check for at least 1 argument - the file name */
  if (nump == 0) 
    tf_error("Usage: $getpat_open(\"file\" [,\"format\"] [,reg_list]);");

  /* Check for not too many arguments */
  if (nump > MAX_ARGS) 
    tf_error("Limit of %d registers", MAX_ARGS);

  /* Check that first arg is a string */
  if (tf_typep(FILE_NAME) != TF_STRING) 
    tf_error("$getpat_open arg 1 must be a quoted file name");


  /* Did the user give an optional mode string? */
  if (tf_typep(FIRST_REG) == TF_STRING)
    first_reg = FIRST_REG + 1;
  else 
    first_reg = FIRST_REG;

  /* Check that other args are reg type */
  for (i=nump; i>=first_reg; i--) {
    if (tf_typep(i) != TF_READWRITE) {
      tf_error("$getpat_open arg %0d must be a register data type", i);
      return FALSE;
    }
    if (tf_sizep(i) > MAX_LINE) {
      tf_error("$getpat_open arg %0d wider than %d", MAX_LINE);
      return FALSE;
    }
  } /* for i */

  return TRUE;
} /* getpat_open_check */


/**********************************************************************
 * $getpat_open call routine
 *********************************************************************/
int getpat_open_call() {
  FILE *inFile;
  int nump = tf_nump();
  int i;
  p_patfile patfile;
  p_sig sig;
  char *str = tf_getcstringp(FILE_NAME);

  if (str == NULL) {
    tf_error("No file name given in call to $getpat_open");
    return 0;
  }

  if ( (inFile = fopen(str,"r")) == NULL) {
    tf_error("$getpat_open cannot open file %s",
               str);
    tf_putp(RETURN_VAL, EOF);
    return(FALSE);
  }

  patfile = (p_patfile) malloc (sizeof(s_patfile) + nump + 1);
  memset((void*)patfile, 0, sizeof(s_patfile) + nump + 1);
  patfile->in_file = inFile;
  patfile->nump = nump;
  patfile->instance = tf_getinstance();
  patfile->line_num = 1;
  patfile->first_read_done = FALSE;

  /* Did the user give an optional mode string? */
  if (tf_typep(FIRST_REG) == TF_STRING) {

    /* Parse the string, check for b,d,h,x,B,D,H,X */
    str = tf_getcstringp(FIRST_REG);
    if (strlen(str) > 2) {
      tf_error("Illegal format \"%s\", legal values are \"b\", \"b \", \"d\", \"x\", \"x \"", 
	       str);
      return 0;
    }
    switch (str[0]) {
    case 'b':
    case 'B':
      patfile->radix = 'b';
      if (str[1] == ' ')
	patfile->separator = ' ';
      break;
      
    case 'd':
    case 'D':
      patfile->radix = 'd';
      patfile->separator = ' ';
      break;
      
    case 'h':
    case 'H':
    case 'x':
    case 'X':
      patfile->radix = 'h';
      if (str[1] == ' ')
	patfile->separator = ' ';
      break;
      
    default:
      tf_error("Illegal format \"%s\", legal values are \"b\", \"b \", \"d\", \"x\", \"x \"", 
	       str);
      return 0;
    } /* switch */

    if ((str[1] != ' ') && (str[1] != '\0')) {
      tf_error("Illegal format \"%s\", legal values are \"b\", \"b \", \"d\", \"x\", \"x \"", 
	       str);
      return 0;
    }

    patfile->first_reg = FIRST_REG + 1;
  }
  else {
    /* No format string, use defaults */
    patfile->radix = 'd';
    patfile->separator = ' ';
    patfile->first_reg = FIRST_REG;
  }



  for (i=patfile->first_reg; i<=nump; i++) {
    sig = (p_sig) malloc (sizeof (s_sig));
    sig->width = tf_sizep(i);
    switch (patfile->radix) {
    case 'b':
      sig->digits = sig->width;
      if (patfile->separator == ' ')
	sprintf(sig->fmt, "%%[01xXzZ?_]"); /* get any length data */
      else
	sprintf(sig->fmt, "%%%0d[01xXzZ?_]", sig->digits); /* fixed length */
      break;

    case 'h':
      sig->digits = (sig->width+3)/4;
      if (patfile->separator == ' ')
	sprintf(sig->fmt, "%%[0-9a-fA-FxXzZ?_]"); /* Get any length data */
      else
	sprintf(sig->fmt, "%%%0d[0-9a-fA-FxXzZ?_]", sig->digits); /* fixed len */
      break;

    case 'd':
      sprintf(sig->fmt, "%%[-+0-9_]"); /* Any width data */
      break;

    default:
      tf_error("Internal error in getpat_open_call: patfile->radix = %c",
	       patfile->radix);
      return 0;
    }

    if (patfile->head_sig != NULL) {
      /* Add to end of list */
      patfile->tail_sig->next_sig = sig;
      patfile->tail_sig = sig;
    }
    else {
      /* First element in the list */
      patfile->tail_sig = sig;
      patfile->head_sig = sig;
    }
    if(gp_verbose)
      io_printf("Arg %0d width=%0d\n", i, sig->width);
  }

  tf_putp(RETURN_VAL, (int) patfile);

  return 0;
} /* getpat_open_call */





/**********************************************************************
 * $getpat_next check routine
 *********************************************************************/
int getpat_next_check(){
  int i,
    nump = tf_nump();

  /* Check for at least a single argument */
  if (nump == 0)
    tf_error("Usage: status = $getpat_next(filenum [, reg_list]);");

  /* Check that first arg is a 32 bit register or integer */
  if ((tf_typep(FILE_ARG) != TF_READWRITE) || (tf_sizep(FILE_ARG) != 32))
    tf_error("$getpat_next arg 1 must be a 32-bit register");

  /* Check that other args are reg type */
  for (i=nump; i>=FILE_ARG+1; i--) {
    if (tf_typep(i) != TF_READWRITE) {
      tf_error("$getpat_next arg %0d must be a register data type", i);
      return(FALSE);
    }
    if (tf_sizep(i) > MAX_LINE) {
      tf_error("$getpat_open arg %0d wider than %d", MAX_LINE);
      return(FALSE);
    }
  } /* for i */


  return 0;
} /* getpat_next_check */



/* ************************************************************************
   $getpat_next routine to strip out the underscores from a number
   Pass in a string with underscores, and this will compress it by
   removing the underscores.
   The string remains in the same place, just missing any '_'
   Also converts lowercase to upper
 */
static void remove_underscores(char *s)
{
  int src, dest;
  for (src = dest = 0; s[src]; ++src) {
    if (s[src] != '_')
      if (islower(s[src]))
	s[dest++] = s[src] - 'a' + 'A';
      else
	s[dest++] = s[src];
  }
  s[dest] = '\0';
} /* remove_underscores */


#define char2int(c) (isdigit(c)? ((c)-'0'): ((c)-'A'+10))
#define int2char(i) ((char)(((i)<10)? ((i)+'0'): ((i)+'A')))

/* ************************************************************************
   $getpat_next routine to truncate a data value to a fixed width
   This is needed because VCS' tf_strdelputp gives a warning if
   you try to write a value that is too large.
   The problem is that for a 5 bit value, 1F is okay, but 20 is not.
   Returns true if data was truncated, false if not.
 */
static int truncate(char *data, int bits, char radix)
{
  int truncated = FALSE;

  if (radix == 'b') {
    /* Binary radix, just truncate */
    int src = strlen(data) - bits,
      dst = 0;
    if (src > 0) {
      truncated = TRUE;
      while (data[src]) 
	data[dst++] = data[src++];
    } // if src
  }
  else {
    /* Must be hex data */
    /* Find the MSB (for this width) in the string */
    int i = strlen(data) - bits / 4 - 1;

    /* Are there any offensive bits? */
    if ( i >= 0 ) {
	int n = 0;
	char digit = data[i];
	static unsigned int mask[] = {0,1,3,7};
        int index = bits % 4;
	/* Should we change the digit, or just delete it? */
	if (index) {
	  unsigned int hexval = char2int(data[i]);
	  data[i] = int2char(hexval & mask[index]);
	  truncated = (digit != data[i]) ||(i > 0);
	}
	else {
	   /* Just skip over the digit all together */
	  i++;
	  truncated = TRUE;
	}

	/* Leftshift to discard any remaining highbits */
	while ( data[i] ) {
	  data[n++] = data[i++];
	}
	data[n] = '\0';
    } // if i >0
  } // else radix

    return truncated;
} /* truncate */


/**********************************************************************
   $getpat_next routine to read a separator from the pattern file
   Always read a separator, even with packed data
   Makes us less likely to puke with badly formatted data
 
   Returns EOF if EOF hit, FALSE if no separator found, TRUE otherwise
*********************************************************************/
int read_separator(p_patfile patfile) {
  int ch, r;
  char line[MAX_LINE];
  bool separator_found = FALSE;

  while (1) {
    ch = fgetc(patfile->in_file);

    switch (ch) {
    case EOF:  return EOF; // End of file
    case ' ':  break;      // Space - do nothing
    case '\t': break;      // Tab - do nothing
    case '\n': 
      patfile->line_num++; // New line
      break;

    case '/':              // Flush comment
      fgets(line, MAX_LINE, patfile->in_file);
      patfile->line_num++;
      break;

    default:
	ungetc(ch, patfile->in_file);
	return separator_found;
	break;

    } // switch (ch)

    /* This is only executed if a whitespace or comment found */
    separator_found = TRUE;

  } // while (1)

} /* read_separator */




/**********************************************************************
 * Function to read just space and tabs, no comments or new-lines
 *********************************************************************/
int read_space_tab(p_patfile patfile) {
  int ch;

  // Only space and tabs allowed - no newlines
  ch = fgetc(patfile->in_file);
  while ((ch == ' ') || (ch == '\t')) {
    ch = fgetc(patfile->in_file);
  }

  switch (ch) {
  case EOF:
    tf_error("EOF detected in the middle of line %d in %s",
	     patfile->line_num, tf_igetcstringp(FILE_NAME, patfile->instance));
    tf_putp(RETURN_VAL, EOF);
    return FALSE;
    break;

  case '\n':
    tf_error("Not enough data values for unpacked data on line %d in %s",
	     patfile->line_num, tf_igetcstringp(FILE_NAME, patfile->instance));
    tf_putp(RETURN_VAL, EOF);
    return FALSE;
    break;

  case '/':
    tf_error("Comment found in the middle of unpacked data on line %d in %s",
	     patfile->line_num, tf_igetcstringp(FILE_NAME, patfile->instance));
    tf_putp(RETURN_VAL, EOF);
    return FALSE;
    break;

  default:
    // Hope this is good data!
    ungetc(ch, patfile->in_file);

  } /* switch ch */

  return TRUE;
} /* read_space_tab */




/**********************************************************************
 * $getpat_next write routine when registers in $getpat_open
 *********************************************************************/
int write_to_open_regs(p_patfile patfile) {
  int arg;
  bool first_sig_on_line;
  p_sig sig;

  first_sig_on_line = TRUE;

  /* Loop through the regs and read data from the file into them */
  sig = patfile->head_sig;
  for (arg=patfile->first_reg; arg<=patfile->nump; arg++) {
    char vector[MAX_LINE];
    int r, i;

    if (sig == NULL) {
      tf_error("Internal error, null signal in getpat_next_call, idx=%d",arg);
      return FALSE;
    }

    // For space delimited files, read a separator between values
    if (!first_sig_on_line && (patfile->separator == ' ')) {
      if (!read_space_tab(patfile))
	return FALSE;
    } /* if first_signal & space separator */

    first_sig_on_line = FALSE;
    patfile->first_read_done = TRUE;

    /* Read the data with the format from getpat_open_call */
    r = fscanf(patfile->in_file, sig->fmt, vector);

    if (gp_verbose) {
      io_printf("Arg %0d with '%s' format, read %s, returned %d\n", 
		arg, sig->fmt, vector, r);
    }

    /* Make sure we read one value */
    if (r != 1) {
      tf_error("Read failed with radix %%%c at line %d in %s", 
	       patfile->radix, patfile->line_num, 
	       tf_igetcstringp(FILE_NAME, patfile->instance));
      tf_putp(RETURN_VAL, EOF);
      return FALSE;
    }

    remove_underscores(vector);

    /* Check if we read enough data from the file */
    if (patfile->radix != 'd') {
      if ((strlen(vector) < sig->digits) &&  (patfile->separator == '\0')) {
	tf_error("Too few digits for packed read at line %d in %s", 
		 patfile->line_num, 
		 tf_igetcstringp(FILE_NAME, patfile->instance));
	tf_putp(RETURN_VAL, EOF);
	return FALSE;
      }
      else {
	if (truncate(vector, sig->width, patfile->radix)) {
	  if (!ignore_truncate)
	    tf_warning("File data truncated for unpacked read arg %d at line %d in %s", 
		       (arg - patfile->first_reg + 1), patfile->line_num, 
		       tf_igetcstringp(FILE_NAME, patfile->instance));
	} /* if truncate */
      } /* if separator */
    } /* if radix != d */


    /* Write the data with a zero-delay non-blocking assignment statement */
    if (!(tf_istrdelputp(arg,sig->width,patfile->radix,vector,0,0,patfile->instance)) ) {
      tf_error("$getpat_next could not write to arg %d at time %s",
	       arg, tf_strgettime() );
      return 0;
    } /* if put */

    sig = sig->next_sig;

  } /* for arg */

  return TRUE;

} /* write_to_open_regs */


/**********************************************************************
 * $getpat_next write routine when registers in $getpat_next
 *********************************************************************/
int write_to_next_regs(p_patfile patfile) {
  int 
    arg,
    nump=tf_nump();
  bool first_sig_on_line;

  first_sig_on_line = TRUE;
  /* Loop through the regs and read data from the file into them */
  for (arg=FILE_ARG+1; arg<=nump; arg++) {
    char fmt[40], vector[MAX_LINE];
    int r,
      digits,
      width = tf_sizep(arg);

    // For space delimited files, read a separator between values
    if (!first_sig_on_line && (patfile->separator == ' ')) {
      if (!read_space_tab(patfile))
	return FALSE;
    } /* if first_signal & space separator */
    first_sig_on_line = FALSE;
    patfile->first_read_done = TRUE;

    digits = 0;
    switch (patfile->radix) {
    case 'b':
      digits = width;
      if (patfile->separator == ' ')
	sprintf(fmt, "%%[01xXzZ?_]");  /* get any length data */
      else
	sprintf(fmt, "%%%0d[01xXzZ?_]", digits); /* fixed length */
      break;

    case 'd':
      sprintf(fmt, "%%[0-9_]"); /* get any length data */
      break;

    case 'h':
      digits = (width+3)/4;
      if (patfile->separator == ' ')
	sprintf(fmt, "%%[0-9a-fA-FxXzZ?_]");  /* get any length data */
      else
	sprintf(fmt, "%%%0d[0-9a-fA-FxXzZ?_]", digits);  /* get any length data */
      break;

    default:
      tf_error("Internal error in write_to_next_regs: patfile->radix = %c",
	       patfile->radix);
      return 0;
    } /* switch radix */

    r = fscanf(patfile->in_file, fmt, vector);

    if (gp_verbose) {
      io_printf("Arg %0d with '%s' format, read %s, returned %d\n", 
		arg, fmt, vector, r);
    }

    /* Make sure we read one value */
    if (r != 1) {
      tf_error("Read failed with radix %%%c at line %d in %s", 
	       patfile->radix, patfile->line_num, 
	       tf_igetcstringp(FILE_NAME, patfile->instance));
      tf_putp(RETURN_VAL, EOF);
      return FALSE;
    }

    remove_underscores(vector);

    /* Check if we read enough data from the file */
    if (patfile->radix != 'd') {
      if ((strlen(vector) < digits) && (patfile->separator == '\0')) {
	tf_error("Too few digits for packed read at line %d in %s", 
		 patfile->line_num, 
		 tf_igetcstringp(FILE_NAME, patfile->instance));
	tf_putp(RETURN_VAL, EOF);
	return FALSE;
      }
      else {
	if (truncate(vector, width, patfile->radix)) {
	  if (!ignore_truncate)
	    tf_warning("File data truncated for unpacked read arg %d at line %d in %s", 
		       (arg - FILE_ARG + 1), patfile->line_num, 
		       tf_igetcstringp(FILE_NAME, patfile->instance));
	} /* if truncate */
      } /* if else separator */
    } /* if radix != d */



    /* Write the data with a zero-delay non-blocking assignment statement */
    if (!(tf_strdelputp(arg,width,patfile->radix,vector,0,0)) ) {
      tf_error("$getpat_next could not write to arg %d at time %s",
	       arg, tf_strgettime() );
      return 0;
    } /* if put */

  } /* for arg */

  return TRUE;

} /* write_to_next_regs */





/**********************************************************************
 * $getpat_next call routine
 *********************************************************************/
int getpat_next_call() {
  int c, r, save_line_num,
    nump = tf_nump(),
    i = tf_getp(FILE_ARG);
  p_patfile patfile;
  p_sig sig;

  if (!i | (i == EOF)) {
    tf_error("$getpat_next passed bad file address %d", i);
    tf_putp(RETURN_VAL, EOF);
    return FALSE;
  }

  patfile = (p_patfile) i;
  if (patfile->in_file == NULL) {
    tf_error("$getpat_next passed closed file");
    tf_putp(RETURN_VAL, EOF);
    return FALSE;
  }


  /* Skip any initial comments or blank lines */
  save_line_num = patfile->line_num;
  r = read_separator(patfile);
  if (r == EOF) {
    tf_putp(RETURN_VAL, EOF);
    return FALSE;
  }

  // A new line must be read at the start of every line of unpacked data
  if ((save_line_num == patfile->line_num) && 
      (patfile->separator == ' ') &&
      patfile->first_read_done) {
    tf_error("Too many data values for unpacked read at line %d in %s",
	     patfile->line_num, tf_igetcstringp(FILE_NAME, patfile->instance));
    tf_putp(RETURN_VAL, EOF);
    return FALSE;
  }


  // If no separator, and this is not the very first thing read...
  if ((r == FALSE) && patfile->first_read_done) {
    tf_error("Too many digits for packed read at line %d in %s",
	     patfile->line_num, tf_igetcstringp(FILE_NAME, patfile->instance));
    tf_putp(RETURN_VAL, EOF);
    return FALSE;
  }
  // patfile->first_read_done is set in write_to_****_regs



  /* Loop through the regs and read data from the file into them */
  sig = patfile->head_sig;
  if (sig) {
    if (nump != FILE_ARG) {
      tf_error("Registers can not be specified in both $getpat_open and $getpat_next - choose only one!");
      tf_putp(RETURN_VAL, EOF);
      return 0;
    }
    if (write_to_open_regs(patfile) == FALSE)
      return 0;
  }
  else {
    if (nump == FILE_ARG) {
      tf_error("Registers must be specified in either $getpat_open or $getpat_next - choose one!");
      tf_putp(RETURN_VAL, EOF);
      return 0;
    }    
    if (write_to_next_regs(patfile) == FALSE)
      return 0;
  }


  tf_putp(RETURN_VAL, patfile->line_num);

  return 0;

} /* getpat_next_call */
/*********************************************************************/






/**********************************************************************
 * $getpat_close check routine
 *********************************************************************/
int getpat_close_check(){
  int i;

  /* Check for a single argument */
  if (tf_nump() != FILE_ARG)
    tf_error("Usage: status = $getpat_close(filenum);");

  /* Check that first arg is a 32 bit register or integer */
  if ((tf_typep(FILE_ARG) != TF_READWRITE) || (tf_sizep(FILE_ARG) != 32))
    tf_error("$getpat_close arg 1 must be a 32-bit register");

  return (0);
} /* getpat_close_check */


/**********************************************************************
 * $getpat_close call routine
 *********************************************************************/
int getpat_close_call() {
  int i = tf_getp(FILE_ARG);
  char c;
  p_patfile patfile;

  if (!i | (i==EOF)) {
    tf_error("$getpat_close passed null file address %0d", i);
    return FALSE;
  }

  patfile = (p_patfile) i;
  if (patfile->in_file == NULL) {
    tf_error("$getpat_close passed closed file");
    return FALSE;
  }

  fclose(patfile->in_file);
  free(patfile);

  return 0;
} /* getpat_close_call */




/**********************************************************************
 * $getpat_options check routine
 *********************************************************************/
int getpat_options_check() {
  int arg, nump = tf_nump();

  if (nump == 0) {
    tf_error ("$getpat_options must have at least one argument");
    return FALSE;
  }

  arg = nump;
  while (arg) {
    char *str;
    int delay;
    if (tf_typep(arg) != tf_string) {
      tf_error("Bad argument %d to $getpat_options, must be string",
		  arg);
      return FALSE;
    }
    else if (tf_sizep(arg) < 3) {
      tf_error("Arguments to $getpat_options must be at least 3 characters, you gave %d",
		  tf_sizep(arg));
      return FALSE;
    }

    arg --;
  } /* while arg */

  return 0;

} /* getpat_options_check */


/**********************************************************************
 * $getpat_options call routine
 *********************************************************************/
int getpat_options_call() {
  int nargs, arg;
  char *str;

  /* Usually I process these arg lists backwards, but here I'm doing first
     to last so later args can reverse earlier - not bloodly likely */
  nargs = tf_nump();
  for (arg = 1; arg <= nargs; arg++) {
      str = tf_getcstringp(arg);

      if (!strncmp(str, "verbose", strlen(str)))
		gp_verbose = 1;
      else if (!strncmp(str, "quiet", strlen(str)))
		gp_verbose = 0;
      else if (!strncmp(str, "ignore_truncate", strlen(str)))
		ignore_truncate = TRUE;
      else if (!strncmp(str, "warn_truncate", strlen(str)))
		ignore_truncate = FALSE;
      else tf_error("Unrecognized argument %s to $getpat_options", str);

  } /* for arg */



  return 0;
} /* getpat_options_call */
/*********************************************************************/


/* Return the size of the return argument of $getpat_open and $getpat_next */
/* Used by both MTI and Verilog-XL */
int getpat_size32() {
  return 32;
}



/************************************************************************
  MTI Support
************************************************************************/

#ifdef MTI
s_tfcell veriusertfs[] = {

  /* stuff for building a shared library for Modelsim 
     Template:
    { usertask|userfunction, data, checktf(), sizetf(), calltf(), misctf(), 
				"$tfname", forwref?, Vtool?, ErrMsg? },
    ***/
    {userfunction, 0, getpat_open_check, getpat_size32, getpat_open_call,  0, "$getpat_open"},
    {userfunction, 0, getpat_next_check, getpat_size32, getpat_next_call,  0, "$getpat_next"},
    {usertask,     0, getpat_close_check, 0,     getpat_close_call, 0, "$getpat_close"},
    {usertask,     0, getpat_options_check, 0,   getpat_options_call, 0, "$getpat_options"},
    {0}                           /*** final entry must be 0 ***/
};


void init_usertfs()
{
  p_tfcell usertf;
  
  for (usertf = veriusertfs;usertf;usertf++) {
    if (usertf->type == 0)
      return;
    mti_RegisterUserTF(usertf);
  }
}
#endif

